/*
deadbit!

by Eli Fulkerson, June 2005
see www.elifulkerson.com for updates.

Originally modified from Ping.c, by Francois Gouget (fgouget@free.fr), which was in turn
based on the ping sample provided with the ICMP libraries (author unknown), the FreeBSD
ping interface, and the FreeBSD traceroute interface.
His website: http://fgouget.free.fr/bing/index-en.shtml


*/

#include <stdio.h>
#include <time.h>
#include <winsock2.h>
#include <ipexport.h>
#include <icmpapi.h>
#include <ws2tcpip.h>
#include <conio.h>
#include <stdbool.h>
#include <stdint.h>

//#include <limits.h>
#include <iphlpapi.h>

#define APP_NAME				"deadbit"
#define APP_VERSION				"0.2"

// predefined constants

#define ICMP_HEADER_SIZE        8		// Don't change this
#define IP_HEADER_SIZE			20		// Don't change this either

// option defaults

#define DEFAULT_OPT_I_VALUE     1000		// Wait 1 second between ICMP packets
#define DEFAULT_OPT_M_VALUE     30		// Same as for TCP
#define DEFAULT_OPT_S_VALUE     56		// Add 56 bytes of data which makes for a 64 bytes ICMP packet.
#define DEFAULT_OPT_W_VALUE     3000L	// Wait for a reply for X milliseconds
#define OPT_P_MAX_SIZE          16		// Patterns are at most 16 bytes long
#define DEFAULT_OPT_R_VALUE		3		// Retry up to three times
#define DEFAULT_PAYLOAD_VALUE	10000	// somewhat low default in order to speed convergence time
#define DEFAULT_OPT_D_VALUE		0		// Debugging display off/on
#define DEFAULT_OPT_REDACT_VALUE	0		// auto-redaction of IP addresses is off

/**
 * A bit field for storing boolean options.
 * Note that in some cases the field is not a boolean and this variable 
 * simply stores whether the option has been used while the actual value is 
 * stored elswhere.
 */
unsigned short options=0;

//#define F_COUNT         0x0001
#define F_INTERVAL      0x0002
#define F_MAXTTL        0x0004
#define F_SLOW          0x0008
#define F_PATTERN       0x0010
#define F_QUIET         0x0020
#define F_IPV6		    0x0040
#define F_TRACE         0x0080
#define F_TIMEOUT       0x0100
#define F_RANDOMFILL    0x0200
#define F_FRAGMENT		0x0400
#define F_PAYLOAD		0x0800

int opt_i = DEFAULT_OPT_I_VALUE;
int opt_m = DEFAULT_OPT_M_VALUE;
int opt_s = DEFAULT_OPT_S_VALUE;
int opt_w = DEFAULT_OPT_W_VALUE;
int opt_r = DEFAULT_OPT_R_VALUE;
int opt_d = DEFAULT_OPT_D_VALUE;

int opt_payload = 1472;

int number_of_icmp_timeouts = 0; // number of times we've waited for an icmp timeout
int opt_timeout_explicit = 0;	// we have explicitly specified a timeout.  don't change what the user wants.

HANDLE hIcmp;
unsigned long count;
unsigned short icmp_data_size=0;
unsigned char* icmp_data=NULL;
unsigned char reply_buffer[100000];
struct icmp_echo_reply* icmp_reply;




// ivp6

//HANDLE hIcmp6;


// On control-c, exit immediately.
BOOL WINAPI ConsoleCtrlHandler(DWORD dwCtrlType) {
    if (dwCtrlType == CTRL_C_EVENT) {
		fprintf(stdout, "\nctrl-c\n");
		exit(1);
	}
	return 0;
}


void formatTimeLeft(long seconds) {
	
	long h = seconds / 3600;
    long m = (seconds - (3600 * h) )  / 60;
	long s = (seconds - (3600 * h)) % 60;
	
	
	if (seconds < 60) {
		fprintf(stdout, "%ds", seconds);
		return;
	}

	if (seconds < 3600) {
		fprintf(stdout, "%dm %ds", m , s);
		return;
	}


	fprintf(stdout, "%dh %dm %ds", h, m, s);

}

/*
 * Converts a string to an IP address. The address may be specified 
 * either as a host name or as the IP address in "dotted" format.
 *
 * @param host_string the host name / address
 * @return the host's IP address if successful, INADDR_NONE otherwise
 */

unsigned int HostString2Addr(char* host_string) {

    unsigned long int host_addr;

	//printf("host string %s host string\n", host_string);
    host_addr = inet_addr(host_string);
    if (host_addr == INADDR_NONE) {
        struct hostent* he;

        he = gethostbyname(host_string);
        if (he == NULL) {
            fprintf(stderr,"%s: unknown host %s\n", APP_NAME, host_string);
        }
		else if (he->h_addrtype != AF_INET) {
            fprintf(stderr,"%s: host %s is not in the AF_INET domain\n", APP_NAME, host_string);
        }
		else {
            memcpy(&host_addr, he->h_addr, sizeof(host_addr));
        }
    }
    return host_addr;
}

/*
 * Returns a pointer to a string suitable for representing the host's
 * address.
 */

char* HostAddr2String(unsigned long int host_addr, int options) {
    static char host_string[15+1+1+64+1+1];
    char* ip_str;



    ip_str = inet_ntoa(*((struct in_addr*)&host_addr));
    //strncpy(host_string,ip_str,sizeof(host_string));
    strncpy_s(host_string, sizeof(host_string), ip_str, sizeof(host_string));

	host_string[sizeof(host_string)-1] = '\0';
    return host_string;
}



/*
 * Builds the data field to be sent with the ICMP packet.
 */
void GenerateData(int size, int first_byte, int last_byte, byte pattern, byte fill_pattern) {
	int i = 0;
	//int r;

    if (icmp_data_size != size) {
        icmp_data_size = size;
        icmp_data=realloc(icmp_data, icmp_data_size);
    }
	
	while (i < icmp_data_size) {
		if (i >= first_byte && i <= last_byte) {
			//fprintf(stderr, "%02hhX", pattern);
			icmp_data[i] = pattern;
		}
		else {
			//fprintf(stderr, " ");
			//fprintf(stderr, "%02hhX", fill_pattern);
			icmp_data[i] = fill_pattern;
		}
		i++;
	}
	
	return;
}

void GenerateData2(int size, int first_byte, int last_byte, byte pattern, byte pattern2, byte fill_pattern) {
	int i = 0;
	//int r;

	if (icmp_data_size != size) {
		icmp_data_size = size;
		icmp_data = realloc(icmp_data, icmp_data_size);
	}

	while (i < icmp_data_size) {
		if (i >= first_byte && i <= last_byte) {
			//fprintf(stderr, "%02hhX", pattern);
			icmp_data[i] = pattern;
			icmp_data[i + 1] = pattern2;
		}
		else {
			//fprintf(stderr, " ");
			//fprintf(stderr, "%02hhX", fill_pattern);
			icmp_data[i] = fill_pattern;
		}
		i++;
		i++;
	}

	return;
}

void GenerateData3(int size, long starting_point) {
	int i = 0;
	
	if (icmp_data_size != size) {
		icmp_data_size = size;
		icmp_data = realloc(icmp_data, icmp_data_size);
	}

	while (i < icmp_data_size) {
		icmp_data[i] = starting_point / 256;
		icmp_data[i + 1] = starting_point % 256;
		starting_point++;
		i++;
		i++;
	}

	return;
}

int doICMP3(int target_addr, int load_size, long starting_point)	{
	int status;

	struct ip_option_information ip_opts;
	PICMP_ECHO_REPLY pEchoReply;
	int maxRetry;

	/* Prepare the IP options */
	memset(&ip_opts, 0, sizeof(ip_opts));
	ip_opts.Ttl = opt_m;
	maxRetry = opt_r;

	/* do not fragment */
	if (!(options & F_FRAGMENT)) {
		ip_opts.Flags = ip_opts.Flags + IP_FLAG_DF;
	}

	while (maxRetry > 0) {

		GenerateData3(load_size, starting_point);

		status = IcmpSendEcho(hIcmp,
			target_addr,
			(LPVOID)icmp_data,
			load_size,
			&ip_opts,
			reply_buffer,
			sizeof(reply_buffer),
			opt_w);

		pEchoReply = (PICMP_ECHO_REPLY)reply_buffer;


		//printf("original status %d", status);
		if (status == 0) {
			status = GetLastError();
		}
		else {
		}

		if (status <= 1) {
			status = pEchoReply->Status;
		}

		if (status == IP_PACKET_TOO_BIG) {
			return IP_PACKET_TOO_BIG;
		}

		if (status == IP_REQ_TIMED_OUT) {
			//printf(".");
			maxRetry--;
			number_of_icmp_timeouts++;
			continue;
		}

		if (status == IP_DEST_NET_UNREACHABLE) {
			//printf("u"); return 0;
		}

		return status;
	}
	return status;
}
    
int doICMP(int target_addr, byte pattern, byte fill_pattern, int load_size, int first_byte, int last_byte, bool doublebyte, byte pattern2)	{
    int status;
    
	struct ip_option_information ip_opts;
	PICMP_ECHO_REPLY pEchoReply;
	int maxRetry;

    /* Prepare the IP options */
    memset(&ip_opts, 0, sizeof(ip_opts));
    ip_opts.Ttl = opt_m;
	maxRetry = opt_r;

	/* do not fragment */
	if (!(options & F_FRAGMENT)) {
		ip_opts.Flags = ip_opts.Flags + IP_FLAG_DF;
	}

	while (maxRetry > 0) {
		
		if (doublebyte) {
			GenerateData2(load_size, first_byte, last_byte, pattern, pattern2, fill_pattern);
		}
		else {
			GenerateData(load_size, first_byte, last_byte, pattern, fill_pattern);
		}

        status = IcmpSendEcho(hIcmp,
                              target_addr,
                              (LPVOID)icmp_data,
				   			  load_size,
                              &ip_opts,
                              reply_buffer,
                              sizeof(reply_buffer),
                              opt_w);
		
		pEchoReply = (PICMP_ECHO_REPLY)reply_buffer;


		if (status == 0) {
			status = GetLastError();
		}
		else {
		}

		if (status <= 1) {
			status=pEchoReply -> Status;
		}

		if (status == IP_PACKET_TOO_BIG) {
			return IP_PACKET_TOO_BIG;
		}

		if (status == IP_REQ_TIMED_OUT) {
			//printf(".");
			maxRetry--;
			number_of_icmp_timeouts++;
			continue;
		}

		if (status == IP_DEST_NET_UNREACHABLE) {
			//printf("u"); return 0;
		}

		return status;
	}

	return status;
}

void output_parameters()
{
	/* do not fragment */
	if (!(options & F_FRAGMENT)) {
		printf("* ICMP Fragmentation is not permitted. *\n");
	}
	else {
		printf("* ICMP Fragmentation is permitted. *\n");
	}

	printf("* Maximum payload is %d bytes. *\n", opt_payload );
}



int doSingleMTU(int target_addr, int display, int payload_size, int first_byte, int last_byte, byte pattern, byte fill_pattern) {

	//int current_check;
	int success = 1;
	int counter = 0;



	int load_size = opt_payload;
	//int first_byte = 0;
	//int last_byte = load_size;

	success = doICMP(target_addr, pattern, fill_pattern, load_size, first_byte, last_byte, false, 0x00);

		

		if (success == 0) {
			// that one worked!  
			fprintf(stdout, "\r passed:  0x%02hhX in position(s) %d through %d of %d bytes.  (other bytes: 0x%02hhX)                                       ", pattern, first_byte, last_byte, load_size, fill_pattern);

		}
		else {
			fprintf(stdout, "\n failed:  0x%02hhX in position(s) %d through %d of %d bytes.  (other bytes: 0x%02hhX)                                        ", pattern, first_byte, last_byte, load_size, fill_pattern);

			if (first_byte == last_byte) {

				int bitsuccess = 0;
				
				for (int k = 0; k < 256; k++) {
					bitsuccess = doICMP(target_addr, k, fill_pattern, load_size, first_byte, last_byte, false, 0x00);

					if (bitsuccess == 0) {
						fprintf(stdout, "\r passed:  0x%02hhX in position(s) %d of %d bytes.  (other bytes: 0x%02hhX)                                        ", k, first_byte, load_size, fill_pattern);
					}
					else {
						fprintf(stdout, "\n failed:  0x%02hhX in position(s) %d of %d bytes.  (other bytes: 0x%02hhX)                                        ", k, first_byte, load_size, fill_pattern);
					}
					Sleep(opt_i);
					//bitsuccess = doICMPBit(target_addr, target_string, load_size);
				}

				// no more recurse
				return;
			}
			int diff = (last_byte - first_byte) / 2;

			Sleep(opt_i);

			// failure, we shall dig deeper
			doSingleMTU(target_addr,  display, payload_size, first_byte, first_byte + diff, pattern, fill_pattern);
			doSingleMTU(target_addr, display, payload_size, first_byte + diff + 1, last_byte, pattern, fill_pattern);
		
		}

	return;
}



/**
 * @param argc number of command-line options
 * @param argv command line options
 * @return 0 if sucessful, 2 for a usage error and 1 otherwise
 */
int main(int argc, char** argv) {
    int retcode;
    WSADATA wsaData;
    char** arg;

	if (argc < 2) {
        goto usage;
	}

    /* Initialisation */
    if (WSAStartup(MAKEWORD(1, 1), &wsaData)) {
        fprintf(stderr, "%s: You must use Winsock 1.1 or compatible\n", APP_NAME);
        goto error;
    }

    hIcmp = IcmpCreateFile();
	//hIcmp6 = Icmp6CreateFile();

    srand((int) time(NULL));
    /* Give a higher priority so that ping has better 
     * chances not to be delayed when measuring the RTT.
     */
    SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);

    icmp_reply = (struct icmp_echo_reply*)reply_buffer;

    /* Install the ^C handler */
    SetConsoleCtrlHandler(&ConsoleCtrlHandler, TRUE);

    /* Parse the parameters and do the pinging and traceing */
    retcode = 0;
    arg = argv + 1;
    //GenerateData(opt_s);
    while (*arg != NULL) {
        if ((*arg)[0] == '-') {
            /* Change some option value */
            switch ((*arg)[1]) {

			case 'h':
			case '?':
				goto usage;
				break;

			case '6':
				/* turn on IPV6 mode */
				options |= F_IPV6;
				break;

            case 'd':
                 /* turns on debugging of ICMP status/failure codes */
				 opt_d++;
				 break;

            case 'i':
                /* Sets the interval between two echo requests. */
                arg++;
                if (sscanf_s(*arg, "%d", &opt_i) != 1)
                    goto usage;
                if (opt_i < 0)
                    opt_i = DEFAULT_OPT_I_VALUE;
                options |= F_INTERVAL;
                break;
            case 'r':
                /* Sets the retry count. */
                arg++;
                if (sscanf_s(*arg, "%d", &opt_r) != 1)
                    goto usage;
                if (opt_r < 0)
                    opt_r = DEFAULT_OPT_R_VALUE;
				opt_timeout_explicit = 1;
                break;

            case 'm':
                /*
                 * Set the maximum payload size
                 */
                arg++;
                if (sscanf_s(*arg, "%d", &opt_payload) != 1)
                    goto usage;
				if (opt_payload <= 92 || opt_payload >= 65280) {
					fprintf(stderr, "Maximum payload must be between 92 and 65280 (inclusive).\n\n");
					goto end;
				}
                if (opt_payload < 0) {
                    opt_payload = DEFAULT_PAYLOAD_VALUE;
				}
                options |= F_PAYLOAD;
                break;

            case 's':
                /*
                 * Skip speed optimizations.  Be slow.
                 */
                options |= F_SLOW;
                break;

			case 'f':
				/*
				 * Switch to nofragment mode.
				 */
				 options |= F_FRAGMENT;
				 break;

            case 'w':
                /*
                 * Sets the time to wait for a response to an Echo Request
                 */
                arg++;
                if (sscanf_s(*arg, "%d", &opt_w) != 1)
                    goto usage;
                if (opt_w < 0)
                    opt_w = DEFAULT_OPT_W_VALUE;
                options |= F_TIMEOUT;
				opt_timeout_explicit = 1;
				
                break;
            case 'z':
                /*
                 * Fills the ICMP packets with random data
                 */
                options = (options & ~F_PATTERN) | F_RANDOMFILL;
                break;
            case 'v':
                /*
                 * Output the version and quit
                 */
				fprintf(stdout, "v%s- compiled: %s %s", APP_VERSION, __DATE__, __TIME__);


                return 0;
                break;
            }
        }
		else
		{
            unsigned long int host_addr;

            /* Get the target host address */
			//testlookup(*arg);
			//host_addr=testlookup(*arg);
			//printf("new %s new \n", host_addr);
            //host_addr=HostString2Addr(*arg);
			//printf("old %s old\n", host_addr);

			//host_addr=HostString2Addr(testlookup(*arg));
			host_addr=HostString2Addr(*arg);

            if (host_addr==INADDR_NONE) {
                retcode=1;
			}
			else {

				//while (counter < 10) {
				if (opt_d == 1) {
					output_parameters();
				}

				int tmp;



					
					fprintf(stdout, "\nTesting:  individual stuck data bits - 1 in a field of 0000...etc  Will bisect if error found...\n");
					tmp = doSingleMTU(host_addr, 1, opt_payload, 0, opt_payload, 0xFF, 0x00);

					
					fprintf(stdout, "\nTesting:  individual stuck data bits - 0 in a field of 1111...etc  Will bisect if error found...\n");
					tmp = doSingleMTU(host_addr, 1, opt_payload, 0, opt_payload, 0x00, 0xFF);

					


					/*
					fprintf(stdout, "\nTest 3 of 6\n");
					fprintf(stdout, "-----------\n");

					fprintf(stdout, "Testing each bit in the checksum field...\n");
					fprintf(stdout, "\nNote:  We are only testing to make sure that 1 and 0 works in each bit of the checksum.  In\n");
					fprintf(stdout, "any given attempt other bits beside the one we are interested in may change as well which may\n");
					fprintf(stdout, "give varying results.\n");
					fprintf(stdout, "\nThis is a blind test - we don't know the checksum, we establish an unknown base and then toggle\n");
					fprintf(stdout, "its bits using binary math.  You will need Wireshark or simlar to see the actual headers that\n");
					fprintf(stdout, "are being sent.\n\n");
						
					*/
					// these should have the same checksum?

					// baseline
					fprintf(stdout, "\nTesting:  each bit in the ICMP checksum\n");
					for (int j = 0; j < 2; j++) {
						tmp = doICMP(host_addr, 0x00, 0x00, opt_payload, 0, 0, true, 0x01);
						if (tmp == 0) {
							fprintf(stdout, "\r passed:  baseline checksum test.  Actual checksum unknown.                                               ");
						}
						else {
							fprintf(stdout, "\n failed:  baseline checksum test.  This was just an arbitrary packet.                                      ");
						}

						for (int i = 0; i < 8; i++) {
							tmp = doICMP(host_addr, 0x00, 0x00, opt_payload, 0, 0, true, pow(2, i) - i);

							long pos = pow(2, i);
							if (tmp == 0) {
								fprintf(stdout, "\r passed:  bit in position 0x%04x of checksum toggled.                                                  ", pos);
							}
							else {
								fprintf(stdout, "\n failed:  bit in position 0x%04x of checksum toggled.                                                    ", pos);
							}
							Sleep(opt_i);
						}

						for (int i = 0; i < 8; i++) {
							// don't need the second -i since the upper byte isn't incremementing
							tmp = doICMP(host_addr, pow(2, i), 0x00, opt_payload, 0, 0, true, 0x00);

							//fprintf(stdout, "%d %d", i, pow(2.0,i));
							long pos = pow(2, i + 8);
							if (tmp == 0) {
								fprintf(stdout, "\r passed:  bit in position 0x%04x of checksum toggled.                                           ", pos);
							}
							else {
								fprintf(stdout, "\n failed:  bit in position 0x%04x of checksum toggled.                                           ", pos);
							}
							Sleep(opt_i);
						}

						/*
						if (j == 0) {
							fprintf(stdout, "\n\n(Running checksum test again in case we wrapped FFFF->0000 the first time.)\n\n");
						}
						*/
						
					}


					//@@TODO - new test zero
					// force every 2 byte from 0x0000 through 0xFFFF through as the being-added term rather than the carry
					fprintf(stdout, "\nTesting:  each possible 16 bit word addend that contributes to the checksum\n");
					bool done = false;
					long k = 0;
					while (!done) {
						tmp = doICMP3(host_addr, opt_payload, k);

						if (tmp == 0) {
							fprintf(stdout, "\r passed:  checksum add register test starting at 0x%04x.                                       ", k);
						}
						else {
							fprintf(stdout, "\n failed:  checksum add register test starting at 0x%04x.                                       ", k);
						} 

						if (k > 65536) {
							done = true;

						}

						k += opt_payload;
					}

					//fprintf(stdout, "\n");

					/*
					fprintf(stdout, "\nTest 4 of 6\n");
					fprintf(stdout, "-----------\n");
					fprintf(stdout, "Trying to force all possible carrys during checkshum calculation...\n");
					*/
					fprintf(stdout, "\nTesting:  all possible checksum carry values\n");
					unsigned long localhost = HostString2Addr("127.0.0.1"); // for the spin

					for (int h = 0; h < (65536 / (opt_payload / 2)) + 1; h++) {
						
						
						// our test.  Theoretically this will express into the carryover register every value from
						// the default from the sequence number through (seqno - opt_payload)
						tmp = doICMP(host_addr, 0xFF, 0xFE, opt_payload, 0, 0, false, 0x00);
						if (tmp == 0) {
							fprintf(stdout, "\r passed:  chksum carry register test #%d of %d.  Actual checksum unknown.", h, (65536 / (opt_payload / 2)) + 1);
							
							fprintf(stdout, "                                              ");
						}
						else {
							fprintf(stdout, "\n failed:  chksum carry register test #%d of %d.  Actual checksum unknown.", h, (65536 / (opt_payload / 2)) + 1);
							
							fprintf(stdout, "                                             ");
						}
						Sleep(opt_i);

						// now we spin the seq number up by opt_payload/2 bytes.  Dividing by two because payload is in bytes and
						// addition is happening in 2 byte words
						for (int i = 0; i < opt_payload / 2; i++) {
							tmp = doICMP(localhost, 0x00, 0x00, 0, 0, 0, true, 0x01);
							// no sleep here, we are pinging localhost so it should be very very fast
						}
						
					}


					int success;
					

					/*
					fprintf(stdout, "\nTest 5 of 6\n");
					fprintf(stdout, "-----------\n");
					fprintf(stdout, "Trying all bytes in each position...\n");
					*/
					fprintf(stdout, "\nTesting:  each data packet consisting of one particular byte over and over\n");
					for (int i= 0; i <= 255; i++) {

						success = doICMP(host_addr, i, 0x00, opt_payload, 0, opt_payload, false, 0x00);
						

						if (success == 0) {
							fprintf(stdout, "\r passed:  0x%02hhX repeating in position(s) %d through %d. ~ ", i, 0, opt_payload);
							
							formatTimeLeft((256 - i)*opt_i / 1000);
							fprintf(stdout, " left                                            ", i, (65536 - i)*opt_i / 1000);
						}
						else {
							fprintf(stdout, "\n failed:  0x%02hhX repeating in position(s) %d through %d.                                        ", i, 0, opt_payload);
						}
						Sleep(opt_i);
					}

					

					/*
					fprintf(stdout, "\n\nTest 6 of 6\n");
					fprintf(stdout, "-----------\n");
					fprintf(stdout, "Incrementing sequence numbers with same payload to test all possible 2 byte ICMP checksums...\n");
					*/
					fprintf(stdout, "\nTesting: incrementing sequence numbers with same payload to test all possible 2 byte ICMP checksums...\n");
					fprintf(stdout, "\n--- THIS TEST IS LAST AND TAKES FOREVER.  ---\n");
					fprintf(stdout, "--- FEEL FREE TO CTRL-C IF YOU DON'T CARE ---\n\n");

					for (long i = 0; i <= 65536; i++) {
						success = doICMP(host_addr, i, 0x00, opt_payload, 0, opt_payload, false, 0x00);
						//}



						if (success == 0) {
							//fprintf(stdout, "passed:  ping #%d of 65536   ~ %ds left...                                             \r", i, (65536-i)*opt_i / 1000);
							fprintf(stdout, "\r passed:  ping #%d of 65536.  ~ ", i) ;
							formatTimeLeft((65536 - i)*opt_i / 1000);
							fprintf(stdout, " left                                             ", i, (65536 - i)*opt_i / 1000);
						}
						else {
							fprintf(stdout, "\n failed:  ping #%d of 65536                                                             ", i);
						}
						Sleep(opt_i);
					}

					fprintf(stdout, "\n");

			}
        }
        arg++;
    }

    /* Do some cleanup */
    IcmpCloseHandle( hIcmp );
    WSACleanup();
    free(icmp_data);

end:
    return retcode;

error:
    retcode = 1;
    goto end;

usage:
	fprintf(stderr, "*****************************************************************\n");
	fprintf(stderr, "  deadbit.exe - written by Eli Fulkerson, Nov 2015             \n");
	fprintf(stderr, "\n");
	fprintf(stderr, "  v%s - %s - %s\n", APP_VERSION, __DATE__, __TIME__);
	fprintf(stderr, "  http://www.elifulkerson.com/projects/deadbit.php for updates. \n");
	fprintf(stderr, "*****************************************************************\n");
	fprintf(stderr,  "\n");
    fprintf(stderr, "Usage: deadbit [-f] [-m SIZE] host \n");
    fprintf(stderr, "Flags:\n");
	fprintf(stderr, "   -f : Allow fragmentation of the packet.\n");
	fprintf(stderr, "   -w : Set the number of milliseconds to wait for a response (default 3000).\n");
	fprintf(stderr, "   -r : Set the maximum number of probe retries on timeout (default = 3).\n");
	fprintf(stderr, "   -i : Set the interval between two echo requests in ms (default 1000).\n");
	fprintf(stderr, "   -m : Sets a payload size to test. (Default is 1472)\n");
    fprintf(stderr, "   -v : Print version info and exit.\n");
	fprintf(stderr, "-h,-? : Print usage information and exit.\n");
	
    retcode = 2;
    goto end;

}